__title__ = "spacy-pytorch-transformers"
__version__ = "0.2.0"
__summary__ = "spaCy pipelines for pre-trained BERT and other transformers"
__uri__ = "https://explosion.ai"
__author__ = "Explosion"
__email__ = "contact@explosion.ai"
__license__ = "MIT"
